#!/usr/bin/perl -w
#
#     demberg.pl [-c|corpus CORPUS | load] [-i|input INPUT] (-s|save) 
#     (-n|ngramtrain TRAINFILE)\n";
#
# example calls:
# demberg.pl -c largewordlist.txt -i tobesegmented.txt -s
# demberg.pl -c load -i tobesegmented.txt -n goodqualitysegmentation.txt
# cat tobesegmented.txt | demberg.pl -c load 
#
# largewordlist.txt should have format "FREQUENCY WORD" for each line,
# use a large corpus or contact me if you want pretrained data.
# tobesegmented.txt should be a list of words "WORD" that you want to
# segment. Can be comprise the same words as largewordlist.
# the option -s allows you to save what you have learnt, so you don't
# need to retrain every time.
# the option -n allows you to specify some good quality segmentation
# that you would like to use for the n-gram model, and is recommended
# for when you only have a very small tobesegmented.txt list of
# words. This -n option allows you to train the ngram model on a
# specified file (e.g. a segmentation of a larger wordlist that you
# generated in an earlier run).
# The program can also read the words that you want to segment from
# the standard input.
# The outputfiles are "INPUTFILE.origSeg" for the naive segmentation
# method and "INPUTFILE.ngramSeg.disambig";
#
# Uses the corpus to learn the morphemes in a language, and then uses these
# morphemes to segment the list of words in the input file.
#
# Copyright (C) 2007 Vera Demberg
# based on program with Copyright (C) 2006  Emily Pitler and Samarth Keshava
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
# MA 02110-1301, USA.


use strict;
use Getopt::Long;

my ($corpus, $input, $save, $ngramtrainfile);
my (@words, %frequencies, $total_count, $cutoff);
my (%front_tree, %back_tree);
my (%prefmap, %sufmap);
my (%stem_candidate);

GetOptions("c|corpus=s" => \$corpus, "i|input=s" => \$input, 
	   "s|save" => \$save, "n|ngramtrain=s" => \$ngramtrainfile) or
    print "Usage: $0 [-corpus CORPUS | load] [-input INPUT] (-save) (-ngramtrain TRAINFILE)\n";

## Read corpus

if ($corpus eq "load"){
    print STDERR "loading data...\n";
    load();
}
else{
open(FIN, $corpus) or die "Could not find corpus: '$corpus'\n";

while(<FIN>) {
    chomp;
    
    my ($num, $word) = /^(\d+) ([^\']*)\'?.*$/;
    chomp($word); 
    $word =~ s/\s+$//; #@V
    next if($num <= 0);
    #$num = 1;
    push @words, $word unless (exists($frequencies{$word}));
    $frequencies{$word} += $num;
    $total_count += $num;
}
close FIN;
}
my @IN;
if ($input){
    open(FIN, $input) or die "Could not find input file: '$input'\n";
    @IN = <FIN>;
    close FIN;
}
else{@IN = <STDIN>;$input = "STDIN";}

###start building up data structures

foreach my $word (@IN) {
    chomp($word); 
    $word =~ s/\s+$//;#(just for robustness)
    push @words, $word unless (exists($frequencies{$word}));
    $frequencies{$word}++;
    $total_count ++;
}

unless($corpus eq "load"){
    $cutoff = (sort { $b <=> $a } values %frequencies)[$#words * .10];
    print STDERR "Read " . ($#words+1) . " distinct words from the corpus.\n";
    print STDERR "Total frequency count: $total_count\n";
    print STDERR "90th percentile frequency: $cutoff\n\n";
    
    print STDERR "Building trees...\n";
    build_tree();
    
    print STDERR "Finding ";
    find_affixes();
}
else{
    print STDERR "updating trees...\n";
    update_trees();
}
if ($save){
    print STDERR "saving...\n";
    save();
}

### Segment words from input file
print STDERR "Segmenting words...\n";
my $origfile = "$input.origSeg";
my $ngramfile = "$input.ngramSeg";
open(ORIGSEG, ">$origfile");
open(NGRAMSEG, ">$ngramfile");
foreach my $word (@IN) {
    chomp $word;
    my $apostrophe = "";
    chomp($word); $word =~ s/\s+$//;#@V
    if($word =~ /^([^\']*)(\'.*)$/) { #can't, don't, we'll... won't? @Vera
	$word = $1;
	$apostrophe = " $2";
    }

    my %break = orig_split_word($word); 
    my %break2 = all_split_word($word, $apostrophe);
    #all_split_word($word, $apostrophe);
    my $sw = "";

    foreach my $pos (0 .. length($word)-1) {
        print ORIGSEG " " if($break{$pos});
        print ORIGSEG substr($word, $pos, 1);
        $sw .=  " " if($break2{$pos});
        $sw .= substr($word, $pos, 1);
    }
    print ORIGSEG $apostrophe, "\n";
    print NGRAMSEG "# $sw";
    print NGRAMSEG $apostrophe, " #\n";
}
close(ORIGSEG);
close(NGRAMSEG);

### train language model
if (not $ngramtrainfile){$ngramtrainfile = $origfile;}
print STDERR "training language model...\n";
system("sh langModel.sh $ngramtrainfile $ngramfile");

#print STDERR "evaluating";
#system("perl evaluation.pl -desired gold-standard.txt -suggested $origfile");
#system("perl evaluation.pl -desired gold-standard.txt -suggested $ngramfile.disambig");

exit(0);


###########################################################################
sub build_tree {
    $front_tree{""} = $total_count;
    $back_tree{""} = $total_count;

    foreach my $word (@words) {
        for(my $size = 1; $size <= length($word); $size++) {
            $front_tree{substr($word, 0, $size)} += $frequencies{$word};
            $back_tree{substr($word, -$size, $size)} += $frequencies{$word};
        }
    }
}
###########################################################################
sub update_trees {
    foreach my $word (@words) {
        for(my $size = 1; $size <= length($word); $size++) {
            $front_tree{substr($word, 0, $size)} += $frequencies{$word};
            $back_tree{substr($word, -$size, $size)} += $frequencies{$word};
        }
    }
}

###########################################################################
sub find_affixes {   
    my $prob_bound = 0.98;
    my (@prefixes, @suffixes);

    ## Find prefixes, suffixes and stems
    print STDERR "prefixes... ";
    my %pref_count = find_prefixes($cutoff, $prob_bound);
    print STDERR "suffixes... ";
    my	%suf_count=find_suffixes($cutoff, \%frequencies, $prob_bound);
    # attention: global variable @stems is filled during find_suffixes.
    print STDERR "more suffixes...\n";
    my %suf_count2 = find_suffixes(0, \%stem_candidate, $prob_bound);
    
    ## Prune 
    @prefixes = prune(\%pref_count);
    @suffixes = prune(\%suf_count);
    my @suffixes2 = prune(\%suf_count2);

    push @suffixes, @suffixes2;
    @suffixes2 = ();

    ## enter data into global hash structure.
    foreach(@suffixes){$sufmap{$_} = 1;}
    foreach(@prefixes){$prefmap{$_}=1;}
}

###########################################################################
sub prune {
    my $pref_count_ref = shift;
    my %pref_count = %$pref_count_ref;
    my @prefixes = sort { length($b) <=> length($a) } keys %pref_count;

    foreach my $prefix (@prefixes) {#delete prefixes that are composed of more than one prefix
        for(my $length = 1; $length < length($prefix); $length++) {
            my $pre1 = substr($prefix, 0, $length);
            my $pre2 = substr($prefix, $length, length($prefix) - $length);

            if(exists($pref_count{$pre1}) and exists($pref_count{$pre2})){
		if ($pref_count{$pre1} > 19 and $pref_count{$pre2} > 19 ){
		    delete $pref_count{$prefix};
		}
	    }
        }
    }    
    @prefixes = grep { $_ and $pref_count{$_} > 19 }
    sort { $pref_count{$b} <=> $pref_count{$a} } keys %pref_count;
    # print "prefixes are:\n",
    #      join("\n", map "$_, $pref_count{$_}", @prefixes), "\n\n";
    return @prefixes;
}

###########################################################################
sub find_prefixes{
    my %pref_count = ();
    my $cutoff = shift;
    my $prob_upper_bound = shift;
    my $prob_lower_bound = $prob_upper_bound;
    foreach  my $word (@words) {
	my $pref_start = 0;
        for(my $pos = 0; $pos < length($word)/2 - 1; $pos++) { 
# restriction on proportional affix length, avoids the risk of cutting a word only 
# ever once (same boundary for prefix and suffix)
            my $pref_size = $pos - $pref_start + 1;
            my $prefix = substr($word, $pref_start, $pref_size);
            my $stem = substr($word, $pos+1);

            ## prob1 = P(first letter of stem | rest of stem)
            ## prob2 = P(last letter of prefix | stem)
            ## (want prob1 to be high and prob2 to be low)
            my $prob1 = $back_tree{$stem} / $back_tree{substr($stem, 1)};
            my $prob2 = $back_tree{substr($word, $pos, 1).$stem} /
                            $back_tree{$stem};
	    # three conditions:
            if(exists($frequencies{$stem}) and
	       $frequencies{$stem} >= $cutoff and
               $prob1 >= $prob_lower_bound and
               $prob2 < $prob_upper_bound) {
		
                $pref_count{$prefix} += 19;
                $pref_start = $pos+1;
            }
            else {
                $pref_count{$prefix}--;
            }
        }
    }
    return %pref_count;
}

###########################################################################
sub find_suffixes{
 my $cutoff = shift;
 my $freq_ref = shift;
 my %local_freqs = %$freq_ref;
 my %suf_count = ();
 my $prob_upper_bound = shift;
 my $prob_lower_bound = $prob_upper_bound;

 my (%suffs, %suffcnt); 

 foreach  my $word (@words) {
	my $suf_start = length($word);

        for(my $pos = length($word)-1; $pos > length($word)/2; $pos--) {
            my $suf_size = $suf_start - $pos;
            my $suffix = substr($word, $pos, $suf_size);
            my $stem = substr($word, 0, $pos);

            ## prob1 = P(last letter of stem | rest of stem)
            ## prob2 = P(first letter of suffix | stem)
            ## (want prob1 to be high and prob2 to be low)
	    my $prob1 = $front_tree{$stem} /
                            $front_tree{substr($word, 0, $pos-1)};
            my $prob2 = $front_tree{substr($word, 0, $pos+1)} /
                            $front_tree{$stem};

            if(exists($local_freqs{$stem}) and
	       $local_freqs{$stem} >= $cutoff and #here: only possible to find stem if stem occurs on its own, but not if only with different affixes. @Vera
               $prob1 >= $prob_lower_bound and
               $prob2 < $prob_upper_bound) {

                $suf_count{$suffix} += 19;
                $suf_start = $pos;
            }
	    #find candidate stems and the suffixes they occur with.
	    # similar to first condition, but register stems that don't occur in corpus.
           elsif($prob1 >= $prob_lower_bound and
		 $prob2 < $prob_upper_bound and 
		 ((not exists($suffs{$stem})) or 
		  not($suffs{$stem} =~ /\s$suffix\s/)) and 
		 not exists($local_freqs{$stem})) {
	       $stem_candidate{$stem}++;
	       $suffs{$stem} .= " $suffix ";
	       $suffcnt{$suffix}++;
            }
            else {
                $suf_count{$suffix}--;
            }
        }
    }
 print STDERR "stems..."unless $cutoff == 0; 
 get_stems(\%suffs, \%suffcnt) unless $cutoff == 0;    
 return %suf_count;
}

###########################################################################
sub get_stems{
    my $suffs_ref = shift;
    my %suffs = %$suffs_ref;
    my $suffcnt_ref = shift;
    my %suffcnt = %$suffcnt_ref;
    my %stem_candidate2;
    my @stems = sort {$stem_candidate{$b} <=> $stem_candidate{$a}} 
    keys %stem_candidate;
    
    open(STEMVAR, ">stemvariation.txt");
    foreach my $s (@stems){
	if ($stem_candidate{$s} > 3){ 
            # only stem candidates with more than one suffix
	    #print $stem_candidate{$s}." $s ";
	    $suffs{$s} =~ s/^\s+// ;
	    my $func =0;
	    my $totalsuffcnt =0;
	    my $comp =0;
	    foreach my $su (split /\s+/, $suffs{$s}){
                #count suffixes that are not compounds.
                if (not(exists($frequencies{$su})) or length($su)<4){
		    $func++; #number of function suffixes
		    $totalsuffcnt += $suffcnt{$su}; # sum of frequencies of suffixes
                }
                else{$comp ++;}
	    }
	    if ($func == 0 ){
		$func =1;
		if ($comp > 3){
		    $totalsuffcnt = 999999;
		    print STEMVAR "$suffs{$s}\n";
                } #good stem since in compounds only
		else {$totalsuffcnt = 1;}
	    }
	    my $ratio = $totalsuffcnt/$func;
	    #print "ratio: ". $ratio."/n";
	    $stem_candidate2{$s} = $ratio;#score for how good a stem is
	    }	
    }
    close(STEMVAR);

    @stems = sort {$stem_candidate2{$b} <=> $stem_candidate2{$a}}
            keys %stem_candidate2;   
    
    my $cutoffstems = int($#stems * 0.3);
    foreach my $s (@stems[$cutoffstems..$#stems]){
        delete $stem_candidate{$s};
    }
}


###########################################################################
sub orig_split_word {
    my $word = shift;
    my (%breaks, %prefs_used, %sufs_used);
    my $original_word = $word;

    my $pref_len_allowed = (1+length($word))/2;#length($word)-3;#
    my $suf_len_allowed = (1+length($word))/2;#length($word)-3;#
    my $pref_start = 0;
    my $suf_start = length($word);
#    my $max = 0;
#find prefixes
    while(1) {
        my $best_prefix;
        my $best_prob = 1;
        my @pref_cands =();
        for (my $i =0;$i< length($word); $i++){
                my $p = substr($word,0,$i);
                if(exists($prefmap{$p})){
                push @pref_cands , $p;}
        }
        foreach my $prefix (@pref_cands) {
            next if($prefs_used{$prefix} or
                    $pref_len_allowed < length($prefix));
            if ($word =~ /^$prefix(.*)$/){
		my $prob = $back_tree{substr($prefix,-1,1).$1} / $back_tree{$1};
		if($prob < $best_prob) {
		    $best_prob = $prob;
		    $best_prefix = $prefix;
		}
	    }
	}
        if($best_prefix and $word =~ /^$best_prefix(.*)$/){
            $word = $1;
            $breaks{$pref_start + length($best_prefix)} = 1;
            $prefs_used{$best_prefix} = 1;
            $pref_start += length($best_prefix);
            $pref_len_allowed -= length($best_prefix);
        }
        else {
            last;
        }
    }
    $word = $original_word;
#find suffixes
    while(1) {
        my $best_suffix;
        my $best_prob = 1;
        my @suff_cands =();
        for (my $i =length($word);$i>0; $i--){
                my $s = substr($word,(length($word)-$i),$i);
                if(exists($sufmap{$s})) {
                push @suff_cands , $s;}
        }

        foreach my $suffix (@suff_cands) {
            next if($suf_len_allowed < length($suffix));
	    if ($word =~ /^(.*)$suffix$/){
                my $prob = $front_tree{$1.substr($suffix, 0, 1)} /
		    $front_tree{$1};
		
                if($prob < $best_prob) {
                    $best_prob = $prob;
                    $best_suffix = $suffix;
                }
	    }
	}
        if($best_suffix and $word =~ /^(.*)$best_suffix$/){
            $word = $1;
            $breaks{$suf_start - length($best_suffix)} = 1;

            $sufs_used{$best_suffix} = 1;
            $suf_start -= length($best_suffix);
            $suf_len_allowed -= length($best_suffix);
        }
        else {
            last;
        }
    }
    return %breaks;
}

###########################################################################

sub all_split_word {
    my $word = shift;
    my $apostrophe = shift;
    my $original_word = $word;
    my %allbreaks;
    my @panalyses = all_pref($word); # generate all prefixations
    my @sanalyses = all_suff($word); # generate all suffixations
    my %solutions = combine_overlap(\@panalyses, \@sanalyses, $original_word);

##### calculate heuristic score
    my $best = -4;
    my $bestres = $original_word;
    print NGRAMSEG "#\n";
    my %greater0;
    my %minus2;
    my %minus3;
    my $gs = 0;
    foreach my $sw (keys %solutions){	
        my @parts = split /\s/, $sw;
        foreach my $part (@parts){
	    unless (exists $prefmap{$part} or 
		    exists ($sufmap{$part}) or 
		    exists ($stem_candidate{$part}) or 
		    exists $frequencies{$part}){
		if (length($part) >0 and length($part)<3){
		    $solutions{$sw} = -3; $minus3{$sw}=1;last;
		}
		else {$solutions{$sw} = -2; $minus2{$sw}=1; last;}
	    }
        }
        $solutions{$sw} = $#parts if $solutions{$sw} > 0;
        $greater0{$sw} =1 if $solutions{$sw} > 0;	
	$gs ++ if $solutions{$sw} > 0;
        if ($solutions{$sw} > $best){$best = $solutions{$sw}; $bestres = $sw;}
    }
##### print
    if ($gs == 0) { 
	foreach my $s (keys %minus2){
	    print NGRAMSEG "# ".$s."$apostrophe #\n";
	}
    }    
    else{
	foreach my $s (keys %greater0) {
	    print NGRAMSEG "# ".$s."$apostrophe #\n";
	} 
	print NGRAMSEG "# ".$bestres."$apostrophe #\n";
    }
    my %breaks = %{getbreaks($bestres)};
    return %breaks;
}

###########################################################################
sub getbreaks{
my $w = shift;
my @l = split /\s+/, $w;
my $p = 0;
my %breaks;
#pop @l;
foreach my $t (@l){
$p += length($t);
$breaks{$p}=1;
}
return \%breaks;
}


###########################################################################
sub readin{
my $file = shift;
open (P,"$file") or die "Could not open $file.";
my  @prefixes;
foreach ( <P>){
chomp;
s/, \d+//g;
push @prefixes, $_;
}
close(P);
return \@prefixes;
}


###########################################################################
sub min{
my $listref = shift;
my @list = @$listref;
my @sortedlist = sort {$a <=> $b} @list;
my $w = shift @sortedlist;
return $w;
}


sub max{
my $listref = shift;
my @list = @$listref;
my @sortedlist = sort {$b <=> $a} @list;
my $w = shift @sortedlist;
return $w;
}


###########################################################################
sub save{
#rausschreiben: frequencies, front_tree, back_tree, prefmap, sufmap, stem_candidate

open(F, ">save.frequencies");
foreach my $w (sort keys %frequencies){print F "$w=>$frequencies{$w}\n";}
close(F);

open(F, ">save.front_tree");
foreach my $w (sort keys %front_tree){print F "$w=>$front_tree{$w}\n";}
close(F);

open(F, ">save.back_tree");
foreach my $w (sort keys %back_tree){print F "$w=>$back_tree{$w}\n";}
close(F);

open(F, ">save.prefmap");
foreach my $w (sort keys %prefmap){print F "$w=>$prefmap{$w}\n";}
close(F);

open(F, ">save.sufmap");
foreach my $w (sort keys %sufmap){print F "$w=>$sufmap{$w}\n";}
close(F);

open(F, ">save.stem_candidate");
foreach my $w (sort keys %stem_candidate){print F "$w=>$stem_candidate{$w}\n";}
close(F);
}

###########################################################################
sub load{
open(F, "save.frequencies") or die "can't open save.frequencies";
while(<F>){/^(.*)=>(.*)$/; $frequencies{$1} = $2;}
close(F);

open(F, "save.front_tree") or die "can't open save.front_tree";
while(<F>){/^(.*)=>(.*)$/; $front_tree{$1} = $2;}
close(F);

open(F, "save.back_tree") or die "can't open save.back_tree";
while(<F>){/^(.*)=>(.*)$/; $back_tree{$1} = $2;}
close(F);

open(F, "save.prefmap") or die "can't open save.prefmap";
while(<F>){/^(.*)=>(.*)$/; $prefmap{$1} = $2;}
close(F);

open(F, "save.sufmap") or die "can't open save.sufmap";
while(<F>){/^(.*)=>(.*)$/; $sufmap{$1} = $2;}
close(F);

open(F, "save.stem_candidate") or die "can't open save.stem_candidate";
while(<F>){/^(.*)=>(.*)$/; $stem_candidate{$1} = $2;}
close(F);
}


######################################################################
sub all_pref{
    my $word = shift;
    my @panalyses;
    my $pref_len_allowed = length($word)-3;#(1+length($word))/2;
    my @preflist = ();
    push @preflist, $word;
    my $best_prob_pref;
    while ($#preflist>=0 ){
        my $z = pop(@preflist);#print "..$z..";
	$word = $z; 
	my $first = "";
	if ($word =~ s/^(.+\s)?(\S+)$/$2/){
	    $first = $1;
	}
        my $best_prefix;
        my $best_prob = 1;my @pref_cands =();
        for (my $i =0;$i< length($word); $i++){
	    my $p = substr($word,0,$i);
	    if(exists ($prefmap{$p})){
                push @pref_cands , $p;}
        }
        foreach my $prefix (@pref_cands) {
	    my $stem = ""; 
            if ($word =~ /^$prefix(.+)$/){$stem = $1;}
            my $prob = $back_tree{substr($prefix,-1,1).$stem} / $back_tree{$stem};
	    if (defined($first)){
		push @preflist, "$first$prefix $stem";
	    }
	    else{push @preflist, "$prefix $stem";}
            if($prob < $best_prob) {
		$best_prob = $prob;
		$best_prefix = $prefix;
            }
        }	
        if($best_prefix and $word =~ /^$best_prefix(.*)$/) {
	    $word = $1;
	    $pref_len_allowed -= length($best_prefix);
	    $best_prob_pref = $best_prob;
        }
        push @panalyses, $z;	
    }
    return @panalyses;
}


######################################################################
sub all_suff{
    my $word = shift;
    my @sanalyses;
    my $best_prob_suff;
    my $suf_len_allowed = length($word)-3;#(1+length($word))/2;
    my @sufflist = ();
    push @sufflist, $word;
    while ($#sufflist>=0 ){
	my @suff_cands =();
        my $best_suffix;
        my $best_prob = 1;
        my $z = pop(@sufflist);
	$word = $z; 
	my $first = "";
	if($word =~ s/^(\S+)(\s?.+)$/$1/){
	    $first = $2;
	}       
        for (my $i =length($word);$i>0; $i--){
	    my $s = substr($word,(length($word)-$i),$i);
	    if(exists($sufmap{$s})) {
                push @suff_cands , $s;}
        }	
        foreach my $suffix (@suff_cands) {
            next if($suf_len_allowed < length($suffix));
	    my $stem = ""; 
	    if ($word =~ /^(.+)$suffix$/){ 
		$stem = $1;
		my $prob = $front_tree{$stem.substr($suffix, 0, 1)} /
		    $front_tree{$stem};
		if (defined ($first)){
		    push @sufflist, "$stem $suffix$first";
		}
		else{push @sufflist, "$stem $suffix";}
		if($prob < $best_prob) {
		    $best_prob = $prob;
		    $best_suffix = $suffix;
		}
	    }
        }	
        if($best_suffix and $word =~ /^(.*)$best_suffix$/) {
            $word = $1;
            $best_prob_suff = $best_prob;
            $suf_len_allowed -= length($best_suffix);
	    
        }
        push @sanalyses, $z;
    }
    return @sanalyses;
}


######################################################################
sub combine_overlap{
    my $paref = shift;
    my $saref = shift;
    my $word = shift;
    my @panalyses = @$paref;
    my @sanalyses = @$saref;
    my %solutions;
    foreach my $s (@sanalyses){	
        foreach my $p (@panalyses){
	    my %allbreaks = %{getbreaks($s)};
	    foreach my $prefboundary (keys %{getbreaks($p)}){
		$allbreaks{$prefboundary} = 1;
	    }
	    my $sw = "";
	    foreach my $pos (0 .. length($word)-1) {
		$sw .=  " " if($allbreaks{$pos});
		$sw .= substr($word, $pos, 1);
	    }	    
	    $solutions{$sw} = 1;
        }
    }
    
    return %solutions;
}
######################################################################
